# Authors: Steve and Clint 
#  Random map stuff was by Clint, who can be reached at (hanclinto at gmail dot com)
# For the Pyweek4 competition, begun March 31st, 2007
#
# This file defines a basic class for holding map data, and creating random maps.
# All maps are generated from a single random number generator seed, such that it's
#  possible to go back to old maps, just by passing that seed in again.

import copy
from random import Random

class coord:  # just a quick way of holding x/y coordinates
	X = 0
	Y = 0
	def __init__(self, x, y):
		self.X = x
		self.Y = y

class tile_style:
	num_styles = 4
	default = 0
	
class tile_type:
	closed, floor = range(2) # these are the two tile types, a sort of enum
	
class tile():
	tile_ = None
	panels = None # for rendering, don't worry about this
	bPanel = None # also for rendering
	style = None
	is_door = None # If the tile is a door, then it can be changed from floor to closed
	in_use = None # This is a flag for us to know if the tile has been assigned or not whilst drawing the map (a needed artifact from porting Clint's original algorithm over)
	coords = None

	def __init__(self):
		self.tile_ = tile_type.closed
		self.coords = coord(0,0)
		self.panels = []
		self.bPanel = False
		self.style = tile_style.default
		self.is_door = False
		self.in_use = False
	
	
class room:
	ul = coord(0,0)
	lr = coord(0,0)
	def __init__(self, ul, lr):
		self.ul = copy.deepcopy(ul)
		self.lr = copy.deepcopy(lr)
		
class gamemap:
	tile_size = 0.0
	map_array = None
	map_width = 0
	map_height = 0

	min_room_size = 5	# Must be an odd number (due to how the room/corridor arangement works)
	max_room_size = 20	# Must be smaller than both MAP_WIDTH and MAP_HEIGHT (oobviously)
	num_rooms = 10		# Can be as large or as small as you want -- if there isn't room for more rooms, it just won't place them.
	room_array = []
	
	start = coord(10, 10)
	stairs = coord(15, 15)
	rand = None
	
	def __init__(self, width, height, size):
		print "map init"
		self.map_array = [tile() for i in xrange(width * height)]
		self.map_width = width
		self.map_height = height
		for y in range(self.map_height):
			for x in range(self.map_width):
				self.map_array[x + self.map_width * y].coords.X = x
				self.map_array[x + self.map_width * y].coords.Y = y
		self.tile_size = size
		self.rand = Random()
		self.clearMap()
		
	def clearMap( self ):
		self.room_array=[]
		self.room_cnt = 0
		
		for x in range(self.map_width):
			for y in range(self.map_height):
				self.setTileType(x, y, tile_type.closed)
				self.getTile(x, y).in_use = False
				self.getTile(x, y).is_door = False
					
	def getTileType(self, X, Y):
#		return copy.deepcopy(self.map_array[X+Y*self.map_width].tile_)
		return copy.deepcopy(self.getTile(X, Y).tile_)
				
	def setTileType(self, X, Y, type_):
#		self.map_array[X+Y*self.map_width].tile_ = copy.deepcopy(type_)
		self.getTile(X, Y).tile_ = copy.deepcopy(type_)
		
	def getTileStyle(self, X, Y):
#		return copy.deepcopy(self.map_array[X+Y*self.map_width].style)
		return copy.deepcopy(getTile(X, Y).style)
			
	def setTileStyle(self, X, Y, style):
#		self.map_array[X+Y*self.map_width].style = copy.deepcopy(style)
		self.getTile(X, Y).style = copy.deepcopy(style)
	
	def getTile(self, X, Y):
		return self.map_array[X+Y*self.map_width]
	
	def setPanel(self,X,Y, bool):
		self.map_array[X+Y*self.map_width].bPanel = copy.copy(bool);
		
	def getPanel(self,X,Y):
		return copy.copy(self.map_array[X+Y*self.map_width].bPanel)
		
	def panelAppend(self, X,Y, panel):
		self.map_array[X+Y*self.map_width].panels.append(panel)
		
	def panelClear(self, X, Y):
		for panel in self.map_array[X+Y*self.map_width].panels:
			panel.remove()
		self.map_array[X+Y*self.map_width].panels = []

	def getTileLocation(self, X, Y):
		vec = vector3df(0,0,0)
		vec.X = X * self.tile_size
		vec.Y = Y * self.tile_size
		return vec
		
	def getStart(self):
		return copy.deepcopy(self.start)
		# this is a "safe" accessor of sorts, it returns a copy instead of the original
		
	def getStairs(self):
		return copy.deepcopy(self.stairs)
	
	def isInRoom(self, X, Y):
		for rum in self.room_array:
			if(X > rum.ul.X + 1 and X < rum.lr.X - 2 and Y > rum.ul.Y + 1 and Y < rum.lr.Y - 2):
				return True
		return False

	def isReallyInRoom(self, X, Y):
		for rum in self.room_array:
			if(X > rum.ul.X and X < rum.lr.X - 1 and Y > rum.ul.Y and Y < rum.lr.Y - 1):
				return True
		return False
			
	def makeRoom(self, ul, lr):
		# Pick a visual style for the room
		style = self.rand.randrange(0, tile_style.num_styles)

		# Add the room into our list so that we can keep track of where it is.
		self.room_array.append(room(ul, lr))

		# Iterate and set the proper wall types		
		for x in range(ul.X, lr.X):
			for y in range(ul.Y, lr.Y):
				if (x == ul.X) or (x == lr.X - 1) or (y == ul.Y) or (y == lr.Y - 1):
					self.setTileType(x, y, tile_type.closed)
				else:
					self.setTileType(x, y, tile_type.floor)
				self.setTileStyle(x, y, style)
				self.getTile(x, y).in_use = True

	def generateCrap(self):
		corner1 = coord(9,9)
		corner2 = coord(16,16)
		self.makeRoom(corner1, corner2)  # makes a single rectangle of open area
				
	# Picks a random seed and creates a random map with it
	def generateRandomMap( self ):
		self.generateMap(None) # Passing in None will automatically use the time for the seed
	
	# Creates a map based on a specified random number generator seed
	def generateMap( self, seed ):
		### Initialization ###
		# Initialize the map
		self.clearMap()
		
		# Initialize our random number generator
		self.rand = Random()
		self.rand.seed(seed)

		### Add rooms ###
		# Loop a maximum number of times
		loopCnt = 0
		while (self.room_cnt < self.num_rooms) and (loopCnt < self.num_rooms * 4): # Try to place rooms up to 4 * self.num_rooms
			# Create new room coordinates
			width = self.rand.randrange(self.min_room_size, self.max_room_size, 2)
			height = self.rand.randrange(self.min_room_size, self.max_room_size, 2)
			left = self.rand.randrange(0, self.map_width - width, 2)
			top = self.rand.randrange(0, self.map_height - height, 2)
			
			ul = coord(left, top)
			lr = coord(left + width, top + height)
			
			# Check to see if new room space is available
			if (self.spaceIsClear(ul, lr)):
				# Add the room
				self.makeRoom(ul, lr)
				self.room_cnt += 1
				
			loopCnt += 1

		# Note: I didn't want corridors to travel through the corners of rooms, so I put all room corners on even numbered coordinates,
		# 	and all corridor corners on odd numbered coordinates -- this has the nice effect of keeping the dungeon looking organized,
		#	as well as preventing collisions between room corners and corridors. It's a bit of a hack, but it works.
		
		### Add corridors ###
		# Shuffle the rooms
		self.rand.shuffle( self.room_array )
		# Connect them each in order to each other
		for cnt in range(self.room_cnt - 1):
			# Choose random X and Y coordinates in each pair of rooms
			x1 = self.rand.randrange( self.room_array[cnt].ul.X + 1, self.room_array[cnt].lr.X, 2 )
			y1 = self.rand.randrange( self.room_array[cnt].ul.Y + 1, self.room_array[cnt].lr.Y, 2 )

			x2 = self.rand.randrange( self.room_array[cnt+1].ul.X + 1, self.room_array[cnt+1].lr.X, 2 )
			y2 = self.rand.randrange( self.room_array[cnt+1].ul.Y + 1, self.room_array[cnt+1].lr.Y, 2 )
			
			c1 = coord(x1, y1)
			c2 = coord(x2, y2)
			
			# Then connect them with a corridor
			self.drawCorridor( c1, c2 )
			
		### Place the up and down stairs ###
		# Pick an arbitrary location in the first room, and make this the down stairs (starting location)
		x = self.rand.randrange( self.room_array[0].ul.X + 1, self.room_array[0].lr.X - 1 )
		y = self.rand.randrange( self.room_array[0].ul.Y + 1, self.room_array[0].lr.Y - 1 )
		
		self.start = coord(x, y)
		
		# Pick an arbitrary location in the last room, and make this the up stairs
		x = self.rand.randrange( self.room_array[self.room_cnt - 1].ul.X + 1, self.room_array[self.room_cnt - 1].lr.X - 1 )
		y = self.rand.randrange( self.room_array[self.room_cnt - 1].ul.Y + 1, self.room_array[self.room_cnt - 1].lr.Y - 1 )

		self.stairs = coord(x, y)
		
	# Draws a corridor between two arbitrary points
	def drawCorridor( self, c1, c2 ):
		# Mole-tunnel from one location to the next
		
		# Start at the beginning
		x = c1.X
		y = c1.Y
		
		while (x != c2.X) or (y != c2.Y):
			if (x < c2.X):
				self.chewHorizTunnel(x, x + 2, y)
				x += 2	
			elif (x > c2.X):
				self.chewHorizTunnel(x, x - 2, y)
				x -= 2
			if (y < c2.Y):
				self.chewVertTunnel(x, y, y + 2)
				y += 2
			elif (y > c2.Y):
				self.chewVertTunnel(x, y, y - 2)
				y -= 2
	
	# A utility function that chews a tunnel horizontally between two points			
	def chewHorizTunnel( self, x1, x2, y ):
		x = x1		
		while (x != x2):
			self.setCorridorTile(x, y)
			
			if (x < x2):
				x += 1
			else:
				x -= 1
		self.setCorridorTile(x, y)

	# A utility function that chews a tunnel vertically between two points
	def chewVertTunnel( self, x, y1, y2 ):
		y = y1
		while (y != y2):
			self.setCorridorTile(x, y)
				
			if (y < y2):
				y += 1
			else:
				y -= 1
		self.setCorridorTile(x, y)
	
	# A utility function that is used to put doors in proper places
	def setCorridorTile( self, x, y ):
		if (self.getTile(x, y).in_use):
			if (self.getTileType(x, y) == tile_type.closed): # If we are trying to set a corridor on a wall, then randomly pick an open or closed door.
				self.getTile(x, y).is_door = True
				#self.setTileType(x, y, self.rand.choice((tile_type.floor, tile_type.closed)))
				self.setTileType(x, y, tile_type.floor)
		else:
			self.setTileType(x, y, tile_type.floor)
			
		self.getTile(x, y).in_use = True		
		
	# A utility function to tell if there is available space to add a room at the requested location
	def spaceIsClear( self, ul, lr ):
		for x in range(ul.X, lr.X):
			for y in range(ul.Y, lr.Y):
				if (self.getTile(x, y).in_use):
					return False # If we hit a tile that is something other than undefined, then we know that the coast isn't clear, and we break

		# If we got this far, then we know that the space is clear
		return True	
	
	# Debug output function to visually display the map
	def dump( self ):
		displayString = ""
		for y in range(self.map_height):
			displayString = ""
			for x in range(self.map_width):
				if (self.getStart().X == x) and (self.getStart().Y == y):
					displayString += ">"
				elif (self.getStairs().X == x) and (self.getStairs().Y == y):
					displayString += "<"
				elif (self.getTile(x, y).in_use):
					if (self.getTile(x, y).is_door):
						if (self.getTileType(x, y) == tile_type.floor):
							displayString += "-"
						else:
							displayString += "+"
					else:
						if (self.getTileType(x, y) == tile_type.floor):
							displayString += "."
						else:
							displayString += "#"
				else:
					displayString += " "

			print displayString


# BEGIN PSEUDO UNIT TEST
#gm = gamemap(80, 30, 1)
#print "\nOur blank map:\n"
#gm.dump()
#print "\n\nAfter adding a room:\n"
#gm.makeRoom( coord(2, 4), coord(7, 11) )
#gm.dump()

#print "\n\nNow our random map 1:\n"
#gm.generateMap(1)
#gm.dump()

#print "\n\nNow our random map 2:\n"
#gm.generateMap(2)
#gm.dump()

#print "\n\nNow a completely random map:\n"
#gm.generateRandomMap()
#gm.dump()

#print "\nDoes it look good? If so, then unit test succeeded."